import { PaymentMethodToken } from '@primer-io/checkout-web'
import Vue from 'vue'
import Component, { createDecorator } from 'vue-class-component'
import {Watch} from 'vue-property-decorator'

type PromiseResolver = {
  resolve: (value: unknown) => void;
  reject: (err: Error) => void;
}


type CallbackMixinOptions = {
  statusPropsName: string;
  functionPropsName: string;
  finishEventName: string;
  eventName: string;
}


export const createCallbackMixin = (options: CallbackMixinOptions) => {
  const MixinProps = Vue.extend({
    props: {
      [options.statusPropsName]: {
        type: [Boolean, Error],
        required: false
      },

      [options.functionPropsName]: {
        type: Function,
        required: false
      }
    },
  });

  let mixin = class CallbackMixin extends MixinProps {
    static __decorators__: any[]

    callbackPromise: any = undefined

    constructor(...args: any) {
      super(...args);
    }

    clearOnTokenizeSuccess() {
      this.callbackPromise = undefined
      this.$emit(options.finishEventName)
    }

    isWaitingForOnTokenizeSuccess() {
      return !!this.callbackPromise
    }

    resolveOnTokenizeSuccess(err: Error) {
      if (!this.isWaitingForOnTokenizeSuccess()) {
        return
      }

      if (err === null || err === undefined) {
        this.callbackPromise?.resolve(undefined)
      } else if (err instanceof Error) {
        this.callbackPromise?.reject(err)
      }

      this.clearOnTokenizeSuccess()
    }

    onTokenizeSuccess(arg: PaymentMethodToken) {
      // Using function props
      const fn =  this[options.functionPropsName]
      if (typeof fn === "function") {
        return fn(arg)
      }

      // Promise to return
      const promise = new Promise((resolve, reject) => {
        this.callbackPromise = {resolve, reject}
      })

      // Create callback to pass to event
      const next = (err: Error) => {
        this.resolveOnTokenizeSuccess(err)
      }

      this.$emit(options.eventName, arg, next)

      return promise
    }

    onTokenizeSuccessStatusChange(tokenizeSuccessStatus: boolean | Error) {
      if (!this.isWaitingForOnTokenizeSuccess()) {
        return
      }

      if (tokenizeSuccessStatus === true) {
        this.callbackPromise?.resolve(undefined)
        this.clearOnTokenizeSuccess()
      } else if (tokenizeSuccessStatus instanceof Error) {
        this.callbackPromise?.reject(tokenizeSuccessStatus)
        this.clearOnTokenizeSuccess()
      }
    }

  }

  mixin.__decorators__ = []
  mixin.__decorators__.push((componentOptions: { watch: any }) => {
    componentOptions.watch ||= Object.create(null)
    componentOptions.watch[options.statusPropsName] ||= [];
    componentOptions.watch[options.statusPropsName].push({ handler: "onTokenizeSuccessStatusChange"  })
  })


  return Component(mixin);
}


